@@ -4,28 +4,32 @@ Fasta file -> Faidx -> Fasta -> FastaRecord -> Sequence
 """
 
 from __future__ import division
+
 import os
+import re
+import string
 import sys
+import warnings
+from collections import namedtuple
+from itertools import islice
+from math import ceil
 from os.path import getmtime
-from six import PY2, PY3, string_types, integer_types
+from threading import Lock
+
+from six import PY2, PY3, integer_types, string_types
 from six.moves import zip_longest
+
 try:
     from collections import OrderedDict
 except ImportError:  #python 2.6
     from ordereddict import OrderedDict
-from collections import namedtuple
-import re
-import string
-import warnings
-from math import ceil
-from threading import Lock
 
 if sys.version_info > (3, ):
     buffer = memoryview
 
 dna_bases = re.compile(r'([ACTGNactgnYRWSKMDVHBXyrwskmdvhbx]+)')
 
-__version__ = '0.5.5.1'
+__version__ = '0.5.9.1'
 
 
 class KeyFunctionError(ValueError):
@@ -123,7 +127,7 @@ class Sequence(object):
         >chr1
         AC
         """
-        if self.start is None or self.end is None:
+        if self.start is None or self.end is None or len(self.seq) == 0:
             correction_factor = 0
         elif len(
                 self.seq
@@ -198,6 +202,13 @@ class Sequence(object):
         """
         return len(self.seq)
 
+    def __eq__(self, other):
+        """
+        >>> Sequence('chr1', 'ACT') == 'ACT'
+        True
+        """
+        return str(self) == str(other)
+
     @property
     def fancy_name(self):
         """ Return the fancy name for the sequence, including start, end, and complementation.
@@ -294,7 +305,7 @@ class IndexRecord(
         return tuple.__getitem__(self, key)
 
     def __str__(self):
-        return "{rlen:n}\t{offset:n}\t{lenc:n}\t{lenb:n}\n".format(
+        return "{rlen:d}\t{offset:d}\t{lenc:d}\t{lenb:d}".format(
             **self._asdict())
 
     def __len__(self):
@@ -449,7 +460,7 @@ class Faidx(object):
     def _index_as_string(self):
         """ Returns the string representation of the index as iterable """
         for k, v in self.index.items():
-            yield '\t'.join([k, str(v)])
+          yield '{k}\t{v}\n'.format(k=k, v=str(v))
 
     def read_fai(self):
         try:
@@ -461,7 +472,7 @@ class Faidx(object):
                     rname, rlen, offset, lenc, lenb = line.split('\t')
                     rlen, offset, lenc, lenb = map(int,
                                                    (rlen, offset, lenc, lenb))
-                    newlines = int(ceil(rlen / lenc) * (lenb - lenc))
+                    newlines = int(ceil(rlen / lenc) * (lenb - lenc)) if lenc else 0
                     bend = offset + newlines + rlen
                     rec = IndexRecord(rlen, offset, lenc, lenb, bend,
                                       prev_bend)
@@ -508,8 +519,8 @@ class Faidx(object):
                     rname = None  # reference sequence name
                     offset = 0  # binary offset of end of current line
                     rlen = 0  # reference character length
-                    blen = None  # binary line length (includes newline)
-                    clen = None  # character line length
+                    blen = 0  # binary line length (includes newline)
+                    clen = 0  # character line length
                     bad_lines = []  # lines > || < than blen
                     thisoffset = offset
                     valid_entry = False
@@ -535,9 +546,9 @@ class Faidx(object):
                                     "Inconsistent line found in >{0} at "
                                     "line {1:n}.".format(
                                         rname, bad_lines[0][0] + 1))
-                            blen = None
+                            blen = 0
                             rlen = 0
-                            clen = None
+                            clen = 0
                             bad_lines = []
                             try:  # must catch empty deflines (actually these might be okay: https://github.com/samtools/htslib/pull/258)
                                 rname = line.rstrip('\n\r')[1:].split()[
@@ -596,7 +607,7 @@ class Faidx(object):
     def write_fai(self):
         with self.lock:
             with open(self.indexname, 'w') as outfile:
-                for line in self._index_as_string:
+                for line in self._index_as_string():
                     outfile.write(line)
 
     def from_buffer(self, start, end):
@@ -648,8 +659,8 @@ class Faidx(object):
 
         # Calculate offset (https://github.com/samtools/htslib/blob/20238f354894775ed22156cdd077bc0d544fa933/faidx.c#L398)
         newlines_before = int(
-            (start0 - 1) / i.lenc * (i.lenb - i.lenc)) if start0 > 0 else 0
-        newlines_to_end = int(end / i.lenc * (i.lenb - i.lenc))
+            (start0 - 1) / i.lenc * (i.lenb - i.lenc)) if start0 > 0 and i.lenc else 0
+        newlines_to_end = int(end / i.lenc * (i.lenb - i.lenc)) if i.lenc else 0
         newlines_inside = newlines_to_end - newlines_before
         seq_blen = newlines_inside + seq_len
         bstart = i.offset + newlines_before + start0
@@ -669,12 +680,15 @@ class Faidx(object):
             else:
                 self.file.seek(bstart)
 
+                # If the requested sequence exceeds len(FastaRecord), return as much as possible
                 if bstart + seq_blen > i.bend and not self.strict_bounds:
                     seq_blen = i.bend - bstart
-
+                # Otherwise it should be safe to read the sequence
                 if seq_blen > 0:
                     seq = self.file.read(seq_blen).decode()
-                elif seq_blen <= 0 and not self.strict_bounds:
+                # If the requested sequence is negative, we will pad the empty string with default_seq.
+                # This was changed to support #155 with strict_bounds=True.
+                elif seq_blen <= 0:
                     seq = ''
 
         if not internals:
@@ -813,7 +827,7 @@ class FastaRecord(object):
             raise
 
     def __iter__(self):
-        """ Construct a line-based iterator that respects the original line lengths. """
+        """ Construct a line-based generator that respects the original line lengths. """
         line_len = self._fa.faidx.index[self.name].lenc
         start = 0
         while True:
@@ -822,11 +836,11 @@ class FastaRecord(object):
                 yield self[start:end]
             else:
                 yield self[start:]
-                raise StopIteration
+                return
             start += line_len
 
     def __reversed__(self):
-        """ Reverse line-based iterator """
+        """ Reverse line-based generator """
         line_len = self._fa.faidx.index[self.name].lenc
         # have to determine last line length
         last_line = len(self) % line_len
@@ -839,7 +853,7 @@ class FastaRecord(object):
                 yield self[start:end][::-1]
             else:
                 yield self[:end][::-1]
-                raise StopIteration
+                return
             if end == len(self):  # first iteration
                 end -= last_line
             else:
@@ -994,13 +1008,9 @@ class Fasta(object):
             sequence_always_upper=sequence_always_upper,
             rebuild=rebuild,
             build_index=build_index)
-        self.keys = self.faidx.index.keys
-        if not self.mutable:
-            self.records = dict(
-                [(rname, FastaRecord(rname, self)) for rname in self.keys()])
-        elif self.mutable:
-            self.records = dict([(rname, MutableFastaRecord(rname, self))
-                                 for rname in self.keys()])
+        
+        _record_constructor = MutableFastaRecord if self.mutable else FastaRecord
+        self.records = OrderedDict([(rname, _record_constructor(rname, self)) for rname in self.faidx.index.keys()])
 
     def __contains__(self, rname):
         """Return True if genome contains record."""
@@ -1009,7 +1019,7 @@ class Fasta(object):
     def __getitem__(self, rname):
         """Return a chromosome by its name, or its numerical index."""
         if isinstance(rname, integer_types):
-            rname = tuple(self.keys())[rname]
+            rname = next(islice(self.records.keys(), rname, None))
         try:
             return self.records[rname]
         except KeyError:
@@ -1019,8 +1029,11 @@ class Fasta(object):
         return 'Fasta("%s")' % (self.filename)
 
     def __iter__(self):
-        for rname in self.keys():
-            yield self[rname]
+        return iter(self.records.values())
+
+    def __len__(self):
+        """Return the cumulative length of all FastaRecords in self.records."""
+        return sum(len(record) for record in self)
 
     def get_seq(self, name, start, end, rc=False):
         """Return a sequence by record name and interval [start, end).
@@ -1057,6 +1070,15 @@ class Fasta(object):
         # len(Sequence.seq) != end - start
         return Sequence(name=name, seq=seq, start=None, end=None)
 
+    def keys(self):
+        return self.records.keys()
+
+    def values(self):
+        return self.records.values()
+
+    def items(self):
+        return self.records.items()
+
     def close(self):
         self.__exit__()
 
@@ -1258,6 +1280,21 @@ def check_bad_lines(rname, bad_lines, i):
     raise RuntimeError("Unhandled exception during fasta indexing at entry " + rname + \
                        "Please report this issue at https://github.com/mdshw5/pyfaidx/issues " + \
                        str(bad_lines))
+    
+def get_valid_filename(s):
+    """
+    From https://github.com/django/django/blob/efc3e32d6d7fb9bb41be73b80c8607b653c1fbd6/django/utils/text.py#L222-L232
+    Return the given string converted to a string that can be used for a clean
+    filename. Remove leading and trailing spaces; convert other spaces to
+    underscores; and remove anything that is not an alphanumeric, dash,
+    underscore, or dot.
+    >>> get_valid_filename("HPV16_144-1.fa")
+    'HPV16_144-1.fa'
+    >>> get_valid_filename("chromosome 6.fa")
+    'chromosome_6.fa'
+    """
+    s = str(s).strip().replace(' ', '_')
+    return re.sub(r'(?u)[^-\w.]', '', s)
 
 
 if __name__ == "__main__":
